# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import threading
import time
import weakref
from abc import ABC, abstractmethod
from typing import Optional, Iterable

from ..types import SocketAddress, AddressPairMap
from ..net import Hub, Channel, Connection, ConnectionDelegate


class ConnectionPool(AddressPairMap[Connection]):

    # Override
    def set(self, item: Optional[Connection],
            remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Connection]:
        # 1. remove cached item
        cached = super().remove(item=item, remote=remote, local=local)
        if cached is not None and cached is not item:
            cached.close()
        # 2. set new item
        old = super().set(item=item, remote=remote, local=local)
        assert old is None, 'should not happen'
        return cached

    # Override
    def remove(self, item: Optional[Connection],
               remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Connection]:
        cached = super().remove(item=item, remote=remote, local=local)
        if cached is not None and cached is not item:
            cached.close()
        if item is not None:
            item.close()
        return cached


class BaseHub(Hub, ABC):

    """
        Maximum Segment Size
        ~~~~~~~~~~~~~~~~~~~~
        Buffer size for receiving package

        MTU        : 1500 bytes (excludes 14 bytes ethernet header & 4 bytes FCS)
        IP header  :   20 bytes
        TCP header :   20 bytes
        UDP header :    8 bytes
    """
    MSS = 1472  # 1500 - 20 - 8

    def __init__(self, delegate: ConnectionDelegate):
        super().__init__()
        self.__delegate = weakref.ref(delegate)
        self.__connection_pool = self._create_connection_pool()
        self.__last_time_drive_connection = time.time()
        self.__lock = threading.Lock()

    # noinspection PyMethodMayBeStatic
    def _create_connection_pool(self):
        return ConnectionPool()

    @property
    def delegate(self) -> ConnectionDelegate:
        return self.__delegate()

    #
    #   Channel
    #

    @abstractmethod
    def _all_channels(self) -> Iterable[Channel]:
        """
        get all channels

        :return: copy of channels
        """
        raise NotImplemented

    @abstractmethod
    def _remove_channel(self, channel: Optional[Channel],
                        remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Channel]:
        """
        remove socket channel

        :param channel: socket channel
        :param remote:  remote address
        :param local:   local address
        """
        raise NotImplemented

    #
    #   Connection
    #

    @abstractmethod
    def _create_connection(self, remote: SocketAddress, local: Optional[SocketAddress]) -> Connection:
        """
        create connection with channel channel & addresses

        :param remote:  remote address
        :param local:   local address
        :return: None on channel not exists
        """
        raise NotImplemented

    def _all_connections(self) -> Iterable[Connection]:
        """ get a copy of all connections """
        return self.__connection_pool.items

    def _remove_connection(self, connection: Optional[Connection],
                           remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Connection]:
        """ remove cached connection """
        return self.__connection_pool.remove(item=connection, remote=remote, local=local)

    def _get_connection(self, remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Connection]:
        """ get cached connection """
        return self.__connection_pool.get(remote=remote, local=local)

    def _set_connection(self, connection: Connection,
                        remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Connection]:
        """ cache connection """
        return self.__connection_pool.set(item=connection, remote=remote, local=local)

    # Override
    def connect(self, remote: SocketAddress, local: Optional[SocketAddress] = None) -> Optional[Connection]:
        # try to get connection
        with self.__lock:
            old = self._get_connection(remote=remote, local=local)
            if old is None:
                # create & cache connection
                conn = self._create_connection(remote=remote, local=local)
                self._set_connection(conn, remote=remote, local=local)
            else:
                conn = old
        if old is None:
            # try to open channel with direction (remote, local)
            conn.start(hub=self)
        return conn

    #
    #   Process
    #

    def _drive_channel(self, channel: Channel) -> bool:
        if not channel.alive:
            # cannot drive closed channel
            return False
        # try to receive
        try:
            data, remote = channel.receive(max_len=self.MSS)
        except socket.error as error:
            remote = channel.remote_address
            local = channel.local_address
            delegate = self.delegate
            if delegate is None or remote is None:
                # UDP channel may not connected
                # so no connection for it
                self._remove_channel(channel=channel, remote=remote, local=local)
            else:
                # remove channel and callback with connection
                conn = self._get_connection(remote=remote, local=local)
                self._remove_channel(channel=channel, remote=remote, local=local)
                if conn is not None:
                    delegate.connection_error(error=error, connection=conn)
            return False
        if remote is None:
            # received nothing
            return False
        else:
            local = channel.local_address
        # get connection for processing received data
        conn = self.connect(remote=remote, local=local)
        if conn is not None:
            conn.received(data=data)
        return True

    def _drive_channels(self, channels: Iterable[Channel]) -> int:
        count = 0
        for sock in channels:
            # drive channel to receive data
            if self._drive_channel(channel=sock):
                count += 1
        return count

    # noinspection PyMethodMayBeStatic
    def _drive_connections(self, connections: Iterable[Connection]):
        now = time.time()
        elapsed = now - self.__last_time_drive_connection
        for conn in connections:
            # drive connection to go on
            conn.tick(now=now, elapsed=elapsed)
            # NOTICE: let the delegate to decide whether close an error connection
            #         or just remove it.
        self.__last_time_drive_connection = now

    def _cleanup_channels(self, channels: Iterable[Channel]):
        for sock in channels:
            if sock.closed:
                # if channel not connected (TCP) and not bound (UDP),
                # means it's closed, remove it from the hub
                self._remove_channel(channel=sock, remote=sock.remote_address, local=sock.local_address)

    def _cleanup_connections(self, connections: Iterable[Connection]):
        # NOTICE: multi connections may share same channel (UDP Hub)
        for conn in connections:
            if conn.closed:
                # if connection closed, remove it from the hub; notice that
                # ActiveConnection can reconnect, it'll be not connected
                # but still open, don't remove it in this situation.
                self._remove_connection(connection=conn, remote=conn.remote_address, local=conn.local_address)

    # Override
    def process(self) -> bool:
        # 1. drive all channels to receive data
        channels = self._all_channels()
        count = self._drive_channels(channels=channels)
        # 2. drive all connections to move on
        connections = self._all_connections()
        self._drive_connections(connections=connections)
        # 3. cleanup closed channels and connections
        self._cleanup_channels(channels=channels)
        self._cleanup_connections(connections=connections)
        return count > 0



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from enum import IntEnum
from typing import Optional, Union

from ..types import SocketAddress
from ..fsm import Processor
from ..net import Connection, ConnectionState
from ..net.state import StateOrder

from .ship import Departure


class DockerStatus(IntEnum):
    """ Docker Status """
    ERROR = -1
    INIT = 0
    PREPARING = 1
    READY = 2


READY_STATUS = [
    StateOrder.READY,
    StateOrder.EXPIRED,
    StateOrder.MAINTAINING
]


def status_from_state(state: Optional[ConnectionState]) -> DockerStatus:
    """ Convert connection state to docker status """
    if state is None:
        return DockerStatus.ERROR
    index = state.index
    if index in READY_STATUS:
        return DockerStatus.READY
    if index == StateOrder.PREPARING:
        return DockerStatus.PREPARING
    if index == StateOrder.ERROR:
        return DockerStatus.ERROR
    return DockerStatus.INIT


class Docker(Processor):
    """
        Star Worker
        ~~~~~~~~~~~

        Processor for Star Ships
    """

    @property
    @abstractmethod
    def closed(self) -> bool:
        """ Connection closed """
        raise NotImplemented

    @property
    @abstractmethod
    def alive(self) -> bool:
        """ Connection alive """
        raise NotImplemented

    @property
    @abstractmethod
    def status(self) -> DockerStatus:
        """ Connection state """
        raise NotImplemented

    @property
    @abstractmethod
    def remote_address(self) -> SocketAddress:
        """ Remote address of connection """
        raise NotImplemented

    @property
    @abstractmethod
    def local_address(self) -> Optional[SocketAddress]:
        """ Local address of connection """
        raise NotImplemented

    @abstractmethod
    def set_connection(self, connection: Optional[Connection]):
        """ Set connection for this docker """
        raise NotImplemented

    @abstractmethod
    def send_data(self, payload: Union[bytes, bytearray]) -> bool:
        """
        Pack data to an outgo ship (with normal priority), and
        append to the waiting queue for sending out

        :param payload: data to be sent
        :return: False on error
        """
        raise NotImplemented

    @abstractmethod
    def send_ship(self, ship: Departure) -> bool:
        """
        Append outgo ship (carrying data package, with priority)
        to the waiting queue for sending out

        :param ship: outgo ship carrying data package/fragment
        :return: False on duplicated
        """
        raise NotImplemented

    @abstractmethod
    def process_received(self, data: bytes):
        """
        Called when received data

        :param data: received data package
        """
        raise NotImplemented

    @abstractmethod
    def heartbeat(self):
        """
        Send 'PING' for keeping connection alive
        """
        raise NotImplemented

    @abstractmethod
    def purge(self, now: float) -> int:
        """ Clear all expired tasks """
        raise NotImplemented

    @abstractmethod
    def close(self):
        """ Close connection for this docker """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2022 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2022 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
import time
import traceback
import weakref
from abc import ABC, abstractmethod
from typing import Set

from .runner import Runner, Daemon


class Ticker(ABC):

    @abstractmethod
    def tick(self, now: float, elapsed: float):
        """
        Drive current thread forward

        :param now:     current time (seconds from Jan 1, 1970 UTC)
        :param elapsed: seconds from previous tick
        """
        raise NotImplemented


class Metronome(Runner):

    # at least wait 1/60 of a second
    MIN_INTERVAL = 1.0/60

    def __init__(self, interval: float):
        super().__init__(interval=interval)
        self.__last_time = 0
        self.__daemon = Daemon(target=self.run)
        self.__lock = threading.Lock()
        self.__tickers = weakref.WeakSet()

    # Override
    def setup(self):
        super().setup()
        self.__last_time = time.time()

    # Override
    def process(self) -> bool:
        tickers = self.tickers
        if len(tickers) == 0:
            # nothing to do now,
            # return False to have a rest ^_^
            return False
        # 1. check time
        now = time.time()
        elapsed = now - self.__last_time
        waiting = self.interval - elapsed
        if waiting < self.MIN_INTERVAL:
            waiting = self.MIN_INTERVAL
        time.sleep(waiting)
        now += waiting
        elapsed += waiting
        # 2. drive tickers
        for item in tickers:
            try:
                item.tick(now=now, elapsed=elapsed)
            except Exception as error:
                print('[Metronome] drive ticker error: %s, %s' % (error, item))
                traceback.print_exc()
        # 3. update last time
        self.__last_time = now
        return True

    @property  # private
    def tickers(self) -> Set[Ticker]:
        with self.__lock:
            return set(self.__tickers)

    def add_ticker(self, ticker: Ticker):
        with self.__lock:
            self.__tickers.add(ticker)

    def remove_ticker(self, ticker: Ticker):
        with self.__lock:
            self.__tickers.remove(ticker)

    def start(self):
        self.__daemon.start()

    def stop(self):
        super().stop()
        self.__daemon.stop()


#
#   Singleton for Prime Metronome
#


class Singleton(object):

    __instances = {}

    def __init__(self, cls):
        self.__cls = cls

    def __call__(self, *args, **kwargs):
        cls = self.__cls
        instance = self.__instances.get(cls, None)
        if instance is None:
            instance = cls(*args, **kwargs)
            self.__instances[cls] = instance
        return instance

    def __getattr__(self, key):
        return getattr(self.__cls, key, None)


@Singleton
class PrimeMetronome:

    def __init__(self):
        metronome = Metronome(interval=Runner.INTERVAL_SLOW)
        metronome.start()
        self.__metronome = metronome

    def add_ticker(self, ticker: Ticker):
        self.__metronome.add_ticker(ticker=ticker)

    def remove_ticker(self, ticker: Ticker):
        self.__metronome.remove_ticker(ticker=ticker)



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import weakref
from abc import ABC, abstractmethod
from typing import Optional, Tuple

from ..types import SocketAddress, AddressPairObject

from ..net.channel import is_blocking, is_closed, is_connected, is_bound
from ..net.channel import bind_socket, connect_socket, disconnect_socket
from ..net import Channel


class SocketReader(ABC):

    @abstractmethod
    def read(self, max_len: int) -> Optional[bytes]:
        """ read data from socket """
        raise NotImplemented

    @abstractmethod
    def receive(self, max_len: int) -> Tuple[Optional[bytes], Optional[SocketAddress]]:
        """ receive data via socket, and return it with remote address """
        raise NotImplemented


class SocketWriter(ABC):

    @abstractmethod
    def write(self, data: bytes) -> int:
        """ write data into socket """
        raise NotImplemented

    @abstractmethod
    def send(self, data: bytes, target: SocketAddress) -> int:
        """ send data via socket with remote address """
        raise NotImplemented


class Controller:
    """ Socket Channel Controller """

    def __init__(self, channel):
        super().__init__()
        self.__channel = weakref.ref(channel)

    @property
    def channel(self):  # -> Optional[BaseChannel]:
        return self.__channel()

    @property
    def remote_address(self) -> Optional[SocketAddress]:
        channel = self.channel
        if isinstance(channel, BaseChannel):
            return channel.remote_address

    @property
    def local_address(self) -> Optional[SocketAddress]:
        channel = self.channel
        if isinstance(channel, BaseChannel):
            return channel.local_address

    @property
    def sock(self) -> Optional[socket.socket]:
        channel = self.channel
        if isinstance(channel, BaseChannel):
            return channel.sock


# noinspection PyAbstractClass
class ChannelReader(Controller, SocketReader, ABC):

    # Override
    def read(self, max_len: int) -> Optional[bytes]:
        sock = self.sock
        if sock is None or is_closed(sock=sock):
            raise ConnectionError('socket closed')
        else:
            return sock.recv(max_len)

    # @abstractmethod  # Override
    # def receive(self, max_len: int) -> Tuple[Optional[bytes], Optional[SocketAddress]]:
    #     """ receive data via socket, and return it with remote address """
    #     raise NotImplemented


# noinspection PyAbstractClass
class ChannelWriter(Controller, SocketWriter, ABC):

    # Override
    def write(self, data: bytes) -> int:
        """ Return the number of bytes sent;
            this may be less than len(data) if the network is busy. """
        # sent = sock.sendall(data)
        sock = self.sock
        if sock is None or is_closed(sock=sock):
            raise ConnectionError('socket closed')
        else:
            return sock.send(data)

    # @abstractmethod  # Override
    # def send(self, data: bytes, target: SocketAddress) -> int:
    #     """ send data via socket with remote address """
    #     raise NotImplemented


class BaseChannel(AddressPairObject, Channel, ABC):

    def __init__(self, sock: socket.socket, remote: Optional[SocketAddress], local: Optional[SocketAddress]):
        super().__init__(remote=remote, local=local)
        # inner socket
        self.__sock = sock
        # create socket reader/writer
        self.__reader = self._create_reader()
        self.__writer = self._create_writer()

    @abstractmethod
    def _create_reader(self) -> SocketReader:
        """ create socket reader """
        raise NotImplemented

    @abstractmethod
    def _create_writer(self) -> SocketWriter:
        """ create socket writer """
        raise NotImplemented

    @property  # protected
    def reader(self) -> SocketReader:
        return self.__reader

    @property  # protected
    def writer(self) -> SocketWriter:
        return self.__writer

    @property
    def sock(self) -> socket.socket:
        return self.__sock

    #
    #   Flags
    #

    @property  # Override
    def blocking(self) -> bool:
        return is_blocking(sock=self.sock)

    @property  # Override
    def closed(self) -> bool:
        return is_closed(sock=self.sock)

    @property  # Override
    def connected(self) -> bool:
        return is_connected(sock=self.sock)

    @property  # Override
    def bound(self) -> bool:
        return is_bound(sock=self.sock)

    @property  # Override
    def alive(self) -> bool:
        return (not self.closed) and (self.connected or self.bound)

    def __str__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s remote="%s" local="%s">\n\t%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.sock, cname, mod)

    def __repr__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s remote="%s" local="%s">\n\t%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.sock, cname, mod)

    # Override
    def configure_blocking(self, blocking: bool):
        sock = self.sock
        sock.setblocking(blocking)
        return sock

    # Override
    def bind(self, address: Optional[SocketAddress] = None,
             host: Optional[str] = '0.0.0.0', port: Optional[int] = 0):
        if address is None:
            if port > 0:
                assert host is not None, 'host should not be empty'
                address = (host, port)
            else:
                address = self._local
                assert address is not None, 'local address not set'
        sock = self.sock
        ok = bind_socket(sock=sock, local=address)
        assert ok, 'failed to bind socket: %s' % str(address)
        self._local = address
        return sock

    # Override
    def connect(self, address: Optional[SocketAddress] = None,
                host: Optional[str] = '127.0.0.1', port: Optional[int] = 0) -> socket.socket:
        if address is None:
            if port > 0:
                assert host is not None, 'host should not be empty'
                address = (host, port)
            else:
                address = self._remote
                assert address is not None, 'remote address not set'
        sock = self.sock
        ok = connect_socket(sock=sock, remote=address)
        assert ok, 'failed to connect socket: %s' % str(address)
        self._remote = address
        return sock

    # Override
    def disconnect(self) -> Optional[socket.socket]:
        sock = self.sock
        ok = disconnect_socket(sock=sock)
        assert ok, 'failed to disconnect socket: %s' % sock
        return sock

    # Override
    def close(self):
        self.disconnect()

    # Override
    def read(self, max_len: int) -> Optional[bytes]:
        try:
            return self.reader.read(max_len=max_len)
        except socket.error as error:
            self.close()
            raise error

    # Override
    def write(self, data: bytes) -> int:
        try:
            return self.writer.write(data=data)
        except socket.error as error:
            self.close()
            raise error

    # Override
    def receive(self, max_len: int) -> Tuple[Optional[bytes], Optional[SocketAddress]]:
        try:
            return self.reader.receive(max_len=max_len)
        except socket.error as error:
            self.close()
            raise error

    # Override
    def send(self, data: bytes, target: SocketAddress) -> int:
        try:
            return self.writer.send(data=data, target=target)
        except socket.error as error:
            self.close()
            raise error



# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
import weakref
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import List, Optional

from .machine import S, C, U, T
from .machine import Transition, State, Machine, Delegate


# noinspection PyAbstractClass
class BaseTransition(Transition[C], ABC):
    """ Transition with the index of target state """

    def __init__(self, target: int):
        super().__init__()
        self.__target = target

    @property
    def target(self) -> int:
        """ target state index """
        return self.__target

    # @abstractmethod  # Override
    # def evaluate(self, ctx: C, now: float) -> bool:
    #     raise NotImplemented


# noinspection PyAbstractClass
class BaseState(State[C, T], ABC):
    """ State with transitions """

    def __init__(self, index: int):
        super().__init__()
        self.__index = index
        self.__transitions: List[Transition[C]] = []

    @property
    def index(self) -> int:
        return self.__index

    def add_transition(self, transition: Transition[C]):
        assert transition not in self.__transitions, 'transition exists: %s' % transition
        self.__transitions.append(transition)

    # Override
    def evaluate(self, ctx: C, now: float) -> Optional[T]:
        for trans in self.__transitions:
            if trans.evaluate(ctx, now=now):
                # OK, get target state from this transition
                return trans

    # @abstractmethod  # Override
    # def on_enter(self, old, ctx: C, now: float):
    #     raise NotImplemented
    #
    # @abstractmethod  # Override
    # def on_exit(self, new, ctx: C, now: float):
    #     raise NotImplemented
    #
    # @abstractmethod  # Override
    # def on_pause(self, ctx: C):
    #     raise NotImplemented
    #
    # @abstractmethod  # Override
    # def on_resume(self, ctx: C):
    #     raise NotImplemented


class MachineStatus(IntEnum):
    """ Machine Status """
    STOPPED = 0
    RUNNING = 1
    PAUSED = 2


class BaseMachine(Machine[C, T, S], ABC):

    def __init__(self):
        super().__init__()
        self.__states: List[S] = []
        self.__current = -1  # current state index
        self.__status = MachineStatus.STOPPED
        self.__delegate_ref: Optional[weakref.ReferenceType] = None

    @property
    def delegate(self) -> Optional[Delegate[C, T, S]]:
        ref = self.__delegate_ref
        if ref is not None:
            return ref()

    @delegate.setter
    def delegate(self, handler: Delegate[C, T, S]):
        self.__delegate_ref = None if handler is None else weakref.ref(handler)

    @property
    @abstractmethod
    def context(self) -> C:
        """ machine itself """
        raise NotImplemented

    #
    #   States
    #
    def add_state(self, state: BaseState[C, T]) -> Optional[S]:
        index = state.index
        assert index >= 0, 'state index error: %d' % index
        count = len(self.__states)
        if index < count:
            # WARNING: return old state that was replaced
            old = self.__states[index]
            self.__states[index] = state
            return old
        # filling empty spaces
        spaces = index - count
        for i in range(spaces):
            self.__states.append(None)
        # append the new state to the tail
        self.__states.append(state)

    def get_state(self, index: int) -> Optional[State[C, T]]:
        return self.__states[index]

    @property  # protected
    def default_state(self) -> State[C, T]:
        return self.__states[0]

    # protected
    def get_target_state(self, transition: BaseTransition[C]) -> State[C, T]:
        # Get target state of this transition
        return self.__states[transition.target]

    @property  # Override
    def current_state(self) -> Optional[State[C, T]]:
        index = self.__current
        if 0 <= index:  # and index < len(self.__states):
            return self.__states[index]

    @current_state.setter  # private
    def current_state(self, state: BaseState[C, T]):
        self.__current = -1 if state is None else state.index

    def __change_state(self, state: Optional[State[C, T]], now: float):
        """
        Exit current state, and enter new state

        :param state:   next state
        :param now:     current time (seconds from Jan 1, 1970 UTC)
        """
        old = self.current_state
        if old == state:
            # print('[FSM] state not change: %s' % state)
            return False
        machine = self.context
        delegate = self.delegate
        #
        #  Events before state changed
        #
        if delegate is not None:
            # prepare for changing current state to the new one,
            # the delegate can get old state via ctx if need
            delegate.enter_state(state, machine, now=now)
        if old is not None:
            old.on_exit(state, machine, now=now)
        #
        #  Change current state
        #
        self.current_state = state
        #
        #  Events after state changed
        #
        if state is not None:
            state.on_enter(old, machine, now=now)
        if delegate is not None:
            # handle after the current state changed,
            # the delegate can get new state via ctx if need
            delegate.exit_state(old, machine, now=now)
        return True

    #
    #   Actions
    #

    # Override
    def start(self):
        now = time.time()
        ok = self.__change_state(state=self.default_state, now=now)
        assert ok, 'failed to change default state'
        self.__status = MachineStatus.RUNNING

    # Override
    def stop(self):
        self.__status = MachineStatus.STOPPED
        now = time.time()
        self.__change_state(state=None, now=now)  # force current state to None

    # Override
    def pause(self):
        now = time.time()
        machine = self.context
        current = self.current_state
        #
        #  Events before state paused
        #
        if current is not None:
            current.on_pause(machine, now=now)
        #
        #  Pause state
        #
        self.__status = MachineStatus.PAUSED
        #
        #  Events after state paused
        #
        delegate = self.delegate
        if delegate is not None:
            delegate.pause_state(current, machine, now=now)

    # Override
    def resume(self):
        now = time.time()
        machine = self.context
        current = self.current_state
        #
        #  Events before state resumed
        #
        delegate = self.delegate
        if delegate is not None:
            delegate.resume_state(current, machine, now=now)
        #
        #  Resume state
        #
        self.__status = MachineStatus.RUNNING
        #
        #  Events after state resumed
        #
        if current is not None:
            current.on_resume(machine, now=now)

    #
    #   Ticker
    #

    # Override
    def tick(self, now: float, elapsed: float):
        machine = self.context
        current = self.current_state
        if current is not None and self.__status == MachineStatus.RUNNING:
            trans = current.evaluate(machine, now=now)
            if trans is not None:
                # assert isinstance(trans, BaseTransition), 'transition error: %s' % trans
                target = self.get_target_state(transition=trans)
                assert target is not None, 'target state error: %s' % trans.target
                self.__change_state(state=target, now=now)



# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
from abc import ABC, abstractmethod
from threading import Thread
from typing import Optional


class Processor(ABC):

    @abstractmethod
    def process(self) -> bool:
        """
        Do the job

        :return: False on nothing to do
        """
        raise NotImplemented


class Handler(ABC):

    @abstractmethod
    def setup(self):
        """ Prepare for Handling """
        raise NotImplemented
    
    @abstractmethod
    def handle(self):
        """ Handling run loop """
        raise NotImplemented

    @abstractmethod
    def finish(self):
        """ Cleanup after handled """
        raise NotImplemented


class Runnable(ABC):

    @abstractmethod
    def run(self):
        """ Run in a thread """
        raise NotImplemented


class Runner(Runnable, Handler, Processor, ABC):
    """
        Runner
        ~~~~~~

        @abstract method:
            - process()
    """

    # Frames Per Second
    # ~~~~~~~~~~~~~~~~~
    # (1) The human eye can process 10-12 still images per second,
    #     and the dynamic compensation function can also deceive us.
    # (2) At a frame rate of 12fps or lower, we can quickly distinguish between
    #     a pile of still images and not animations.
    # (3) Once the playback rate (frames per second) of the images reaches 16-24 fps,
    #     our brain will assume that these images are a continuously moving scene
    #     and will appear like the effect of a movie.
    # (4) At 24fps, there is a feeling of 'motion blur',
    #     while at 60fps, the image is the smoothest and cleanest.
    INTERVAL_SLOW = 1.0/10
    INTERVAL_NORMAL = 1.0/25
    INTERVAL_FAST = 1.0/60

    def __init__(self, interval: float):
        super().__init__()
        assert interval > 0, 'interval error: %s' % interval
        self.__interval = interval
        self.__running = False

    @property
    def interval(self) -> float:
        return self.__interval

    @property
    def running(self) -> bool:
        return self.__running

    def stop(self):
        self.__running = False

    # Override
    def run(self):
        self.setup()
        try:
            self.handle()
        finally:
            self.finish()

    # Override
    def setup(self):
        self.__running = True

    # Override
    def handle(self):
        while self.running:
            if not self.process():
                # if nothing to do now, return False here
                # to let the thread have a rest.
                self._idle()

    # Override
    def finish(self):
        self.__running = False

    def _idle(self):
        time.sleep(self.interval)

    @abstractmethod
    def process(self) -> bool:
        raise NotImplemented


class Daemon:
    """
        Daemon Thread
        ~~~~~~~~~~~~~
        The main thread won't wait the daemon threads exit
        when daemonic = True
    """

    def __init__(self, target, daemonic: bool = True):
        super().__init__()
        self.__target = target  # Callable
        self.__daemon = daemonic
        self.__thread: Optional[Thread] = None

    @property
    def alive(self) -> bool:
        thr = self.__thread
        if thr is not None:
            return thr.is_alive()

    def start(self):
        self.__force_stop()
        thr = Thread(target=self.__target, daemon=self.__daemon)
        thr.start()
        self.__thread = thr

    def stop(self):
        self.__force_stop()

    def __force_stop(self):
        thr = self.__thread
        if thr is not None:
            self.__thread = None
            self._join(thr=thr)

    def _join(self, thr: Thread):
        # Waits at most seconds for this thread to die.
        # A timeout of 0 means to wait forever.
        self.join(thr=thr, timeout=1.0)

    @classmethod
    def join(cls, thr: Thread, timeout: float = None):
        try:
            thr.join(timeout=timeout)
        except RuntimeError as error:
            print('[ERROR] failed to join thread: %s, timeout: %d' % (error, timeout))
            # traceback.print_exc()



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
import weakref
from abc import ABC
from typing import Optional, Any, List, Dict, MutableMapping

from .port import ShipStatus
from .port import Arrival, Departure


# noinspection PyAbstractClass
class DepartureShip(Departure, ABC):

    # Departure task will be expired after 2 minutes
    # if no response received.
    EXPIRES = 120  # seconds

    # Important departure task will be retried 2 times
    # if response timeout.
    RETRIES = 2

    def __init__(self, priority: int = 0, max_tries: int = None):  # max_tries = 1 + RETRIES
        super().__init__()
        # task priority, smaller is faster
        self.__priority = priority
        # expired time (timestamp in seconds)
        self.__expired = 0
        # how many times to try sending
        self.__tries = (1 + self.RETRIES) if max_tries is None else max_tries

    @property
    def priority(self) -> int:
        return self.__priority

    # Override
    def touch(self, now: float):
        assert self.__tries > 0, 'touch error, tries=%d' % self.__tries
        # decrease counter
        self.__tries -= 1
        # update retried time
        self.__expired = now + self.EXPIRES

    # Override
    def get_status(self, now: float) -> ShipStatus:
        fragments = self.fragments
        if fragments is None or len(fragments) == 0:
            return ShipStatus.DONE
        elif self.__expired == 0:
            return ShipStatus.NEW
        # elif not self.is_important:
        #     return ShipStatus.DONE
        elif now < self.__expired:
            return ShipStatus.WAITING
        elif self.__tries > 0:
            return ShipStatus.TIMEOUT
        else:
            return ShipStatus.FAILED


class DepartureHall:
    """ Memory cache for Departures """

    def __init__(self):
        super().__init__()
        # all departure ships
        self.__all_departures = weakref.WeakSet()
        # new ships waiting to send out
        self.__new_departures: List[Departure] = []
        # ships waiting for responses
        self.__fleets: Dict[int, List[Departure]] = {}  # priority => List[Departure]
        self.__priorities: List[int] = []
        # index
        self.__map: MutableMapping[Any, Departure] = weakref.WeakValueDictionary()  # SN => ship
        self.__finished_times: Dict[Any, float] = {}                                # SN => timestamp
        self.__departure_level: Dict[Any, int] = {}                                 # SN => priority

    def add_departure(self, ship: Departure) -> bool:
        """
        Add outgoing ship to the waiting queue

        :param ship: departure task
        :return: false on duplicated
        """
        # 1. check duplicated
        if ship in self.__all_departures:
            return False
        else:
            self.__all_departures.add(ship)
        # 2. insert to the sorted queue
        priority = ship.priority
        index = len(self.__new_departures)
        while index > 0:
            index -= 1
            if self.__new_departures[index].priority <= priority:
                # take the place before first ship
                # which priority is greater than this one.
                index += 1  # insert after
                break
        self.__new_departures.insert(index, ship)
        return True

    def check_response(self, ship: Arrival) -> Optional[Departure]:
        """
        Check response from income ship

        :param ship: incoming ship with SN
        :return finished task
        """
        sn = ship.sn
        assert sn is not None, 'Ship SN not found: %s' % ship
        # check whether this task has already finished
        timestamp = self.__finished_times.get(sn)
        if timestamp is not None and timestamp > 0:
            # task already finished
            return None
        # check departure task
        outgo = self.__map.get(sn, None)
        if outgo is not None and outgo.check_response(ship=ship):
            # all fragments sent, departure task finished
            # remove it and clear mapping when SN exists
            self.__remove_ship(ship=outgo, sn=sn)
            # mark finished time
            self.__finished_times[sn] = time.time()
            return outgo

    def __remove_ship(self, ship: Departure, sn):
        priority = self.__departure_level.get(sn)
        fleet = self.__fleets.get(priority)
        if fleet is not None and ship in fleet:
            fleet.remove(ship)
            # remove array when empty
            if len(fleet) == 0:
                self.__fleets.pop(priority, None)
        # remove mapping by SN
        self.__map.pop(sn, None)
        self.__departure_level.pop(sn, None)
        self.__all_departures.discard(ship)

    def next_departure(self, now: float) -> Optional[Departure]:
        """
        Get next new/timeout task

        :param now: current time
        :return departure task
        """
        # task.__expired == 0
        task = self.__next_new_departure(now=now)
        if task is None:
            # task.__expired < now
            task = self.__next_timeout_departure(now=now)
        return task

    def __next_new_departure(self, now: float) -> Optional[Departure]:
        if len(self.__new_departures) > 0:
            # get first ship
            outgo = self.__new_departures.pop(0)
            sn = outgo.sn
            if outgo.is_important and sn is not None:
                # this task needs response
                # choose an array with priority
                priority = outgo.priority
                self.__insert_ship(ship=outgo, priority=priority, sn=sn)
                # build index for it
                self.__map[sn] = outgo
            else:
                # disposable ship needs no response,
                # remove it immediately
                self.__all_departures.discard(outgo)
            outgo.touch(now=now)
            return outgo

    def __insert_ship(self, ship: Departure, priority: int, sn):
        fleet = self.__fleets.get(priority)
        if fleet is None:
            # create new array for this priority
            fleet = []
            self.__fleets[priority] = fleet
            # insert the priority in a sorted list
            self.__insert_priority(priority=priority)
        # append to the tail, and build index for it
        fleet.append(ship)
        self.__departure_level[sn] = priority

    def __insert_priority(self, priority: int) -> bool:
        index = 0
        for value in self.__priorities:
            if value == priority:
                # duplicated
                return False
            elif value > priority:
                # got it
                break
            else:
                # current value is smaller than the new value,
                # keep going
                index += 1
        # insert new value before the bigger one
        self.__priorities.insert(index, priority)
        return True

    def __next_timeout_departure(self, now: float) -> Optional[Departure]:
        # no need to copy the list being changed because it will return immediately
        priorities = self.__priorities  # list(self.__priorities)
        for prior in priorities:
            # 1. get tasks with priority
            fleet = self.__fleets.get(prior)
            if fleet is None:
                continue
            # 2. seeking timeout task in this priority
            departures = fleet  # list(fleet)
            for ship in departures:
                sn = ship.sn
                status = ship.get_status(now=now)
                if status == ShipStatus.TIMEOUT:
                    # response timeout, needs retry now.
                    # move to next priority
                    fleet.remove(ship)
                    self.__insert_ship(ship=ship, priority=(prior + 1), sn=sn)
                    # update expired time;
                    ship.touch(now=now)
                    return ship
                elif status == ShipStatus.FAILED:
                    # try too many times and still missing response,
                    # task failed, remove this ship.
                    fleet.remove(ship)
                    # remove mapping by SN
                    self.__map.pop(sn, None)
                    self.__departure_level.pop(sn, None)
                    self.__all_departures.discard(ship)
                    return ship

    def purge(self, now: float = 0) -> int:
        """ Clear all expired tasks """
        if now <= 0:
            now = time.time()
        count = 0
        # 1. seeking finished tasks
        priorities = list(self.__priorities)
        for prior in priorities:
            fleet = self.__fleets.get(prior)
            if fleet is None:
                # this priority is empty
                self.__priorities.remove(prior)
                continue
            departures = list(fleet)
            for ship in departures:
                if ship.get_status(now=now) == ShipStatus.DONE:
                    # task done, remove if from memory cache
                    sn = ship.sn
                    assert sn is not None, 'Ship SN should not be empty here'
                    fleet.remove(ship)
                    # remove mapping by SN
                    self.__map.pop(sn, None)
                    self.__departure_level.pop(sn, None)
                    self.__all_departures.discard(ship)
                    # mark finished time
                    self.__finished_times[sn] = now
                    count += 1
            # remove array when empty
            if len(fleet) == 0:
                self.__fleets.pop(prior, None)
                self.__priorities.remove(prior)
        # 2. seeking neglected finished times
        ago = now - 3600
        keys = set(self.__finished_times.keys())
        for sn in keys:
            when = self.__finished_times.get(sn)
            if when is None or when < ago:
                # long time ago
                self.__finished_times.pop(sn, None)
        return count



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
from typing import Optional

from .arrival import Arrival, ArrivalHall
from .departure import Departure, DepartureHall


"""
    Star Dock
    ~~~~~~~~~

    Parking Star Ships
"""


class Dock:

    def __init__(self):
        super().__init__()
        # memory caches
        self.__arrival_hall = self._create_arrival_hall()
        self.__departure_hall = self._create_departure_hall()

    # noinspection PyMethodMayBeStatic
    def _create_arrival_hall(self) -> ArrivalHall:
        """ Override for user-customized hall """
        return ArrivalHall()

    # noinspection PyMethodMayBeStatic
    def _create_departure_hall(self) -> DepartureHall:
        """ Override for user-customized hall """
        return DepartureHall()

    def assemble_arrival(self, ship: Arrival) -> Optional[Arrival]:
        """
        Check income ship for completed package

        :param ship: received ship carrying data package/fragment
        :return a ship carrying completed data package
        """
        # check fragment from income ship,
        # return a ship with completed package if all fragments received
        return self.__arrival_hall.assemble_arrival(ship=ship)

    def add_departure(self, ship: Departure) -> bool:
        """
        Add outgoing ship to the waiting queue

        :param ship: departure task
        :return False on duplicated
        """
        return self.__departure_hall.add_departure(ship=ship)

    def check_response(self, ship: Arrival) -> Optional[Departure]:
        """
        Check response from the income ship

        :param ship: income ship with SN
        :return finished departure task
        """
        # check departure tasks with SN
        # remove package/fragment if matched (check page index for fragments too)
        return self.__departure_hall.check_response(ship=ship)

    def next_departure(self, now: float) -> Optional[Departure]:
        """
        Get next new/timeout task

        :param now: current timestamp
        :return departure task
        """
        # this will be remove from the queue,
        # if needs retry, the caller should append it back
        return self.__departure_hall.next_departure(now=now)

    def purge(self, now: float = 0):
        """ Clear all expired tasks """
        count = 0
        count += self.__arrival_hall.purge(now=now)
        count += self.__departure_hall.purge(now=now)
        return count


class LockedDock(Dock):

    def __init__(self):
        super().__init__()
        self.__lock = threading.Lock()
        # purge
        self.__next_purge_time = 0

    # Override
    def assemble_arrival(self, ship: Arrival) -> Optional[Arrival]:
        with self.__lock:
            return super().assemble_arrival(ship=ship)

    # Override
    def add_departure(self, ship: Departure) -> bool:
        with self.__lock:
            return super().add_departure(ship=ship)

    # Override
    def check_response(self, ship: Arrival) -> Optional[Departure]:
        with self.__lock:
            return super().check_response(ship=ship)

    # Override
    def next_departure(self, now: float) -> Optional[Departure]:
        with self.__lock:
            return super().next_departure(now=now)

    # Override
    def purge(self, now: float = 0) -> int:
        if now < self.__next_purge_time:
            return -1
        else:
            # next purge after half a minute
            self.__next_purge_time = now + 30
        with self.__lock:
            return super().purge(now=now)



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Union

from ..fsm import Context, BaseTransition, BaseState, BaseMachine

from .connection import Connection


"""
    Finite States:

            //===============\\          (Start)          //=============\\
            ||               || ------------------------> ||             ||
            ||    Default    ||                           ||  Preparing  ||
            ||               || <------------------------ ||             ||
            \\===============//         (Timeout)         \\=============//
                                                              |       |
            //===============\\                               |       |
            ||               || <-----------------------------+       |
            ||     Error     ||          (Error)                 (Connected
            ||               || <-----------------------------+   or bound)
            \\===============//                               |       |
                A       A                                     |       |
                |       |            //===========\\          |       |
                (Error) +----------- ||           ||          |       |
                |                    ||  Expired  || <--------+       |
                |       +----------> ||           ||          |       |
                |       |            \\===========//          |       |
                |       (Timeout)           |         (Timeout)       |
                |       |                   |                 |       V
            //===============\\     (Sent)  |             //=============\\
            ||               || <-----------+             ||             ||
            ||  Maintaining  ||                           ||    Ready    ||
            ||               || ------------------------> ||             ||
            \\===============//       (Received)          \\=============//
"""


class StateMachine(BaseMachine, Context):
    """ Connection State Machine """

    def __init__(self, connection: Connection):
        super().__init__()
        self.__conn_ref = weakref.ref(connection)
        # init states
        builder = self._create_state_builder()
        self.add_state(state=builder.get_default_state())
        self.add_state(state=builder.get_preparing_state())
        self.add_state(state=builder.get_ready_state())
        self.add_state(state=builder.get_expired_state())
        self.add_state(state=builder.get_maintaining_state())
        self.add_state(state=builder.get_error_state())

    # noinspection PyMethodMayBeStatic
    def _create_state_builder(self):
        return StateBuilder(transition_builder=TransitionBuilder())

    @property  # Override
    def context(self) -> Context:
        return self

    @property
    def connection(self) -> Connection:
        return self.__conn_ref()


class StateOrder(IntEnum):
    """ Connection State Order """
    INIT = 0  # default
    PREPARING = 1
    READY = 2
    MAINTAINING = 3
    EXPIRED = 4
    ERROR = 5


class StateTransition(BaseTransition[StateMachine], ABC):
    """ Connection State Transition """

    def __init__(self, target: Union[int, StateOrder]):
        if isinstance(target, StateOrder):
            target = target.value
        super().__init__(target=target)

    @abstractmethod  # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        raise NotImplemented


class ConnectionState(BaseState[StateMachine, StateTransition]):
    """
        Connection State
        ~~~~~~~~~~~~~~~~

        Defined for indicating connection status

            DEFAULT     - 'initialized', or sent timeout
            PREPARING   - connecting or binding
            READY       - got response recently
            EXPIRED     - long time, needs maintaining (still connected)
            MAINTAINING - sent 'PING', waiting for response
            ERROR       - long long time no response, connection lost
    """

    def __init__(self, order: StateOrder):
        super().__init__(index=order.value)
        self.__name = str(order)
        self.__time = 0  # enter time

    @property
    def name(self) -> str:
        return self.__name

    @property
    def enter_time(self) -> float:
        return self.__time

    def __str__(self) -> str:
        return self.__name

    def __repr__(self) -> str:
        return self.__name

    def __eq__(self, other) -> bool:
        if isinstance(other, ConnectionState):
            if self is other:
                return True
            return self.index == other.index
        elif isinstance(other, StateOrder):
            return self.index == other.value
        elif isinstance(other, int):
            return self.index == other
        elif isinstance(other, str):
            return self.__name == other
        else:
            return False

    def __ne__(self, other) -> bool:
        if isinstance(other, ConnectionState):
            if self is other:
                return False
            return self.index != other.index
        elif isinstance(other, StateOrder):
            return self.index != other.value
        elif isinstance(other, int):
            return self.index != other
        elif isinstance(other, str):
            return self.__name != other
        else:
            return True

    # Override
    def on_enter(self, old, ctx: StateMachine, now: float):
        self.__time = now

    # Override
    def on_exit(self, new, ctx: StateMachine, now: float):
        self.__time = 0

    # Override
    def on_pause(self, ctx: StateMachine, now: float):
        pass

    # Override
    def on_resume(self, ctx: StateMachine, now: float):
        pass


class TimedConnection(ABC):

    @property
    @abstractmethod
    def last_sent_time(self) -> float:
        raise NotImplemented

    @property
    @abstractmethod
    def last_received_time(self) -> float:
        raise NotImplemented

    @abstractmethod
    def is_sent_recently(self, now: float) -> bool:
        raise NotImplemented

    @abstractmethod
    def is_received_recently(self, now: float) -> bool:
        raise NotImplemented

    @abstractmethod
    def is_not_received_long_time_ago(self, now: float) -> bool:
        raise NotImplemented


#
#   Builders
#

class StateBuilder:

    def __init__(self, transition_builder):
        super().__init__()
        self.__builder: TransitionBuilder = transition_builder

    # Connection not started yet
    def get_default_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.INIT)
        # Default -> Preparing
        state.add_transition(transition=builder.get_default_preparing_transition())
        return state

    # Connection started, preparing to connect/bind
    def get_preparing_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.PREPARING)
        # Preparing -> Ready
        state.add_transition(transition=builder.get_preparing_ready_transition())
        # Preparing -> Default
        state.add_transition(transition=builder.get_preparing_default_transition())
        return state

    # Normal state of connection
    def get_ready_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.READY)
        # Ready -> Expired
        state.add_transition(transition=builder.get_ready_expired_transition())
        # Ready -> Error
        state.add_transition(transition=builder.get_ready_error_transition())
        return state

    # Long time no response, need maintaining
    def get_expired_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.EXPIRED)
        # Expired -> Maintaining
        state.add_transition(transition=builder.get_expired_maintaining_transition())
        # Expired -> Error
        state.add_transition(transition=builder.get_expired_error_transition())
        return state

    # Heartbeat sent, waiting response
    def get_maintaining_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.MAINTAINING)
        # Maintaining -> Ready
        state.add_transition(transition=builder.get_maintaining_ready_transition())
        # Maintaining -> Expired
        state.add_transition(transition=builder.get_maintaining_expired_transition())
        # Maintaining -> Error
        state.add_transition(transition=builder.get_maintaining_error_transition())
        return state

    # Connection lost
    def get_error_state(self) -> ConnectionState:
        builder = self.__builder
        state = ConnectionState(order=StateOrder.ERROR)
        # Error -> Default
        state.add_transition(transition=builder.get_error_default_transition())
        return state


class TransitionBuilder:

    # noinspection PyMethodMayBeStatic
    def get_default_preparing_transition(self):
        return DefaultPreparingTransition(target=StateOrder.PREPARING)

    # Preparing

    # noinspection PyMethodMayBeStatic
    def get_preparing_ready_transition(self):
        return PreparingReadyTransition(target=StateOrder.READY)

    # noinspection PyMethodMayBeStatic
    def get_preparing_default_transition(self):
        return PreparingDefaultTransition(target=StateOrder.INIT)

    # Ready

    # noinspection PyMethodMayBeStatic
    def get_ready_expired_transition(self):
        return ReadyExpiredTransition(target=StateOrder.EXPIRED)

    # noinspection PyMethodMayBeStatic
    def get_ready_error_transition(self):
        return ReadyErrorTransition(target=StateOrder.ERROR)

    # Expired

    # noinspection PyMethodMayBeStatic
    def get_expired_maintaining_transition(self):
        return ExpiredMaintainingTransition(target=StateOrder.MAINTAINING)

    # noinspection PyMethodMayBeStatic
    def get_expired_error_transition(self):
        return ExpiredErrorTransition(target=StateOrder.ERROR)

    # Maintaining

    # noinspection PyMethodMayBeStatic
    def get_maintaining_ready_transition(self):
        return MaintainingReadyTransition(target=StateOrder.READY)

    # noinspection PyMethodMayBeStatic
    def get_maintaining_expired_transition(self):
        return MaintainingExpiredTransition(target=StateOrder.EXPIRED)

    # noinspection PyMethodMayBeStatic
    def get_maintaining_error_transition(self):
        return MaintainingErrorTransition(target=StateOrder.ERROR)

    # noinspection PyMethodMayBeStatic
    def get_error_default_transition(self):
        return ErrorDefaultTransition(target=StateOrder.INIT)


#
#   Transitions
#

class DefaultPreparingTransition(StateTransition):
    """ Default -> Preparing """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        # connection started? change state to 'preparing'
        return not (conn is None or conn.closed)


class PreparingReadyTransition(StateTransition):
    """ Preparing -> Ready """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        # connected or bound, change state to 'ready'
        return conn is not None and conn.alive


class PreparingDefaultTransition(StateTransition):
    """ Preparing -> Default """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        # connection stopped, change state to 'not_connect'
        return conn is None or conn.closed


class ReadyExpiredTransition(StateTransition):
    """ Ready -> Expired """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return False
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection still alive, but
        # long time no response, change state to 'maintain_expired'
        return not conn.is_received_recently(now=now)


class ReadyErrorTransition(StateTransition):
    """ Ready -> Error """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        # connection lost, change state to 'error
        return conn is None or not conn.alive


class ExpiredMaintainingTransition(StateTransition):
    """ Expired -> Maintaining """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return False
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection still alive, and
        # sent recently, change state to 'maintaining'
        return conn.is_sent_recently(now=now)


class ExpiredErrorTransition(StateTransition):
    """ Expired -> Error """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return True
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection lost, or
        # long time no response, change state to 'error'
        return conn.is_not_received_long_time_ago(now=now)


class MaintainingReadyTransition(StateTransition):
    """ Maintaining -> Ready """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return False
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection still alive, and
        # received recently, change state to 'ready'
        return conn.is_received_recently(now=now)


class MaintainingExpiredTransition(StateTransition):
    """ Maintaining -> Expired """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return False
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection still alive, but
        # long time no sending, change state to 'maintain_expired'
        return not conn.is_sent_recently(now=now)


class MaintainingErrorTransition(StateTransition):
    """ Maintaining -> Error """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        # connection lost, change state to 'error'
        if conn is None or not conn.alive:
            return True
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection lost, or
        # long time no response, change state to 'error'
        return conn.is_not_received_long_time_ago(now=now)


class ErrorDefaultTransition(StateTransition):
    """ Error -> Default """

    # Override
    def evaluate(self, ctx: StateMachine, now: float) -> bool:
        conn = ctx.connection
        if conn is None or not conn.alive:
            return False
        assert isinstance(conn, TimedConnection), 'connection error: %s' % conn
        # connection still alive, and
        # can receive data during this state
        current = ctx.current_state
        assert isinstance(current, ConnectionState), 'connection state error: %s' % current
        enter = current.enter_time
        assert enter > 0, 'should not happen'
        return enter < conn.last_received_time



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import traceback
from abc import ABC, abstractmethod
from typing import Optional, Tuple

from ..types import SocketAddress


class Channel(ABC):

    @property
    @abstractmethod
    def closed(self) -> bool:
        """ not is_open() """
        raise NotImplemented

    @property
    @abstractmethod
    def bound(self) -> bool:
        """ is_bound() """
        raise NotImplemented

    @property
    @abstractmethod
    def alive(self) -> bool:
        """ is_opened() and (is_connected() or is_bound()) """
        raise NotImplemented

    @abstractmethod
    def close(self):
        """ Close the channel """
        raise NotImplemented

    #
    #   Byte Channel
    #

    @abstractmethod
    def read(self, max_len: int) -> Optional[bytes]:
        """
        Reads a sequence of bytes from this channel into the given buffer.

        :param max_len: max buffer length
        :return: The number of bytes read, possibly zero,
                 or -1 if the channel has reached end-of-stream
        :raise: socket.error
        """
        raise NotImplemented

    @abstractmethod
    def write(self, data: bytes) -> int:
        """
        Writes a sequence of bytes to this channel from the given buffer.

        :param data: outgo data
        :return: The number of bytes written, possibly zero
        :raise: socket.error
        """
        raise NotImplemented

    #
    #   Selectable Channel
    #

    @abstractmethod
    def configure_blocking(self, blocking: bool):
        """
        Adjusts this channel's blocking mode.

        :param blocking: If True then this channel will be placed in blocking mode;
                         if False then it will be placed non-blocking mode
        :raise: socket.error
        """
        raise NotImplemented

    @property
    @abstractmethod
    def blocking(self) -> bool:
        """ is_blocking() """
        raise NotImplemented

    #
    #   Network Channel
    #

    @abstractmethod
    def bind(self, address: Optional[SocketAddress] = None,
             host: Optional[str] = '0.0.0.0', port: Optional[int] = 0):
        """
        Binds the channel's socket to a local address (host, port).

        :param address: local address
        :param host:    local host
        :param port:    local port
        :return: bound socket
        :raise: socket.error
        """
        raise NotImplemented

    @property
    @abstractmethod
    def local_address(self) -> Optional[SocketAddress]:  # (str, int)
        """
        Returns the socket address that this channel's socket is bound to.

        :return: The socket address that the socket is bound to,
                 or None if the channel's socket is not bound
        """
        raise NotImplemented

    #
    #   Socket/Datagram Channel
    #

    @property
    @abstractmethod
    def connected(self) -> bool:
        """ is_connected() """
        raise NotImplemented

    @abstractmethod
    def connect(self, address: Optional[SocketAddress] = None,
                host: Optional[str] = '127.0.0.1', port: Optional[int] = 0) -> socket.socket:
        """
        Connects this channel's socket.

        :param address: remote address
        :param host:    remote host
        :param port:    remote port
        :return: connected socket
        :raise: socket.error
        """
        raise NotImplemented

    @property
    @abstractmethod
    def remote_address(self) -> Optional[SocketAddress]:  # (str, int)
        """
        Returns the remote address to which this channel's socket is connected.

        :return: The remote address; None if the channel's socket is not connected
        """
        raise NotImplemented

    #
    #   Datagram Channel
    #

    @abstractmethod
    def disconnect(self) -> Optional[socket.socket]:
        """
        Disconnects this channel's socket.

        :return: inner socket
        :raise: socket.error
        """
        raise NotImplemented

    @abstractmethod
    def receive(self, max_len: int) -> Tuple[Optional[bytes], Optional[SocketAddress]]:
        """
        Receives a data package via this channel.

        :param max_len: max buffer length
        :return: received data and remote address
        :raise: socket.error
        """
        raise NotImplemented

    @abstractmethod
    def send(self, data: bytes, target: SocketAddress) -> int:
        """
        Sends a data package via this channel.

        :param data:   outgo data package
        :param target: remote address
        :return: The number of bytes sent, which will be either the number
                 of bytes that were remaining in the source buffer when this
                 method was invoked or, if this channel is non-blocking, may be
                 zero if there was insufficient room for the datagram in the
                 underlying output buffer
        :raise: socket.error
        """
        raise NotImplemented


def get_local_address(sock: socket.socket) -> Optional[SocketAddress]:
    try:
        return sock.getsockname()
    except socket.error:
        # print('[NET] failed to get local address: %s' % error)
        return None


def get_remote_address(sock: socket.socket) -> Optional[SocketAddress]:
    try:
        return sock.getpeername()
    except socket.error:
        # print('[NET] failed to get remote address: %s' % error)
        return None


def is_blocking(sock: socket.socket) -> bool:
    try:
        return sock.getblocking()
    except socket.error:
        # print('[NET] failed to get blocking: %s' % error)
        return False


def is_connected(sock: socket.socket) -> bool:
    return get_remote_address(sock=sock) is not None


def is_bound(sock: socket.socket) -> bool:
    return get_local_address(sock=sock) is not None


def is_closed(sock: socket.socket) -> bool:
    return getattr(sock, '_closed', False)


def bind_socket(sock: socket.socket, local: SocketAddress) -> bool:
    """ Bind to local address """
    try:
        sock.bind(local)
        return is_bound(sock=sock)
    except socket.error as error:
        print('[Socket] cannot bind to: %s, socket: %s, %s' % (local, sock, error))
        traceback.print_exc()
        return False


def connect_socket(sock: socket.socket, remote: SocketAddress) -> bool:
    """ Connect to remote address """
    try:
        sock.connect(remote)
        return is_bound(sock=sock)
    except socket.error as error:
        print('[Socket] cannot connect to: %s, socket: %s, %s' % (remote, sock, error))
        traceback.print_exc()
        return False


def disconnect_socket(sock: socket.socket) -> bool:
    """ Close socket """
    if not is_connected(sock=sock):
        return True
    try:
        # TODO: check for UDP socket
        # sock.shutdown(socket.SHUT_RDWR)
        sock.close()
        return is_closed(sock=sock)
    except socket.error as error:
        print('[Socket] cannot close socket: %s, %s' % (sock, error))
        traceback.print_exc()
        return False



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, List

"""
    Star Ship
    ~~~~~~~~~
    
    Container carrying data package
"""


class DeparturePriority(IntEnum):
    """ Departure Ship Priority """

    URGENT = -1
    NORMAL = 0
    SLOWER = 1


class ShipStatus(IntEnum):
    """ Ship Status """
    #
    #  Departure Status
    #
    NEW = 0x00      # not try yet
    WAITING = 0x01  # sent, waiting for responses
    TIMEOUT = 0x02  # waiting to send again
    DONE = 0x03     # all fragments responded (or no need respond)
    FAILED = 0x04   # tried 3 times and missed response(s)
    #
    #  Arrival Status
    #
    ASSEMBLING = 0x10  # waiting for more fragments
    EXPIRED = 0x11     # failed to received all fragments


class Ship(ABC):

    @property
    @abstractmethod
    def sn(self) -> Any:
        """
        Get ID for this ship

        :return: SN
        """
        raise NotImplemented

    @abstractmethod
    def touch(self, now: float):
        """
        Update expired time

        :param now: current time
        """
        raise NotImplemented

    @abstractmethod
    def get_status(self, now: float) -> ShipStatus:
        """
        Check ship state

        :param now: current time
        :return: current status
        """
        raise NotImplemented


class Arrival(Ship):
    """ Incoming Ship """

    @abstractmethod
    def assemble(self, ship):  # -> Optional[Arrival]:
        """
        Data package can be sent as separated batches

        :param ship: arrival ship carrying with data package/fragment
        :return new ship carrying the whole data package
        """
        raise NotImplemented


class Departure(Ship):
    """ Outgoing Ship """

    @property
    @abstractmethod
    def priority(self) -> int:
        """ Task priority, default is 0, smaller is faster """
        raise NotImplemented

    @property
    @abstractmethod
    def fragments(self) -> List[bytes]:
        """
        Get fragments to sent

        :return remaining (separated) data package
        """
        raise NotImplemented

    @abstractmethod
    def check_response(self, ship: Arrival) -> bool:
        """
        The received ship may carried a response for the departure
        if all fragments responded, means this task is finished.

        :param ship: income ship carrying response
        :return True on task finished
        """
        raise NotImplemented

    #
    #   task states
    #

    @property
    @abstractmethod
    def is_important(self) -> bool:
        """
        Whether needs to wait for responses

        :return false for disposable
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union

from ..types import SocketAddress
from ..fsm import Processor

from .ship import Departure


class Gate(Processor, ABC):
    """
        Star Gate
        ~~~~~~~~~
    """

    @abstractmethod
    def send_data(self, payload: Union[bytes, bytearray],
                  remote: SocketAddress, local: Optional[SocketAddress]) -> bool:
        """
        Pack data to an outgo ship (with normal priority), and
        append to the waiting queue of docker for remote address

        :param payload: data to be sent
        :param remote: remote address
        :param local:  local address
        :return: False on error
        """
        raise NotImplemented

    @abstractmethod
    def send_ship(self, ship: Departure,
                  remote: SocketAddress, local: Optional[SocketAddress]) -> bool:
        """
        Append outgo ship (carrying data package, with priority)
        to the waiting queue of docker for remote address

        :param ship:   departure ship
        :param remote: remote address
        :param local:  local address
        :return: False on error
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .net import Hub, Channel
from .net import Connection, ConnectionDelegate, ConnectionState

from .socket import BaseChannel, ChannelReader, ChannelWriter
from .socket import BaseHub, BaseConnection, ActiveConnection

from .port import Ship, Arrival, Departure
from .port import ShipStatus, DeparturePriority
from .port import Docker, DockerStatus, DockerDelegate
from .port import Gate

from .arrival import ArrivalShip, ArrivalHall
from .departure import DepartureShip, DepartureHall
from .dock import Dock, LockedDock
from .stardocker import StarDocker
from .stargate import StarGate

name = "StarTrek"

__author__ = 'Albert Moky'

__all__ = [

    'Hub', 'Channel',
    'Connection', 'ConnectionDelegate', 'ConnectionState',

    'BaseChannel', 'ChannelReader', 'ChannelWriter',
    'BaseHub', 'BaseConnection', 'ActiveConnection',

    'Ship', 'Arrival', 'Departure',
    'ShipStatus', 'DeparturePriority',
    'Docker', 'DockerStatus', 'DockerDelegate',
    'Gate',

    'ArrivalShip', 'ArrivalHall', 'DepartureShip', 'DepartureHall',
    'Dock', 'LockedDock',
    'StarDocker', 'StarGate',
]



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
import weakref
from abc import ABC, abstractmethod
from typing import Optional, List

from .types import SocketAddress, AddressPairObject
from .net import Connection
from .port import Arrival, Departure, ShipStatus
from .port import Docker, DockerStatus, DockerDelegate
from .port.docker import status_from_state

from .dock import Dock, LockedDock


class StarDocker(AddressPairObject, Docker, ABC):
    """
        Star Docker
        ~~~~~~~~~~~

        @abstract methods:
            - heartbeat()
            - _get_arrivals(data)
            - _check_arrival(ship)
    """

    def __init__(self, remote: SocketAddress, local: Optional[SocketAddress]):
        super().__init__(remote=remote, local=local)
        self.__dock = self._create_dock()
        self.__delegate_ref = None
        self.__conn_ref = None
        self.__closed = None
        # remaining data to be sent
        self.__last_outgo: Optional[Departure] = None
        self.__last_fragments: List[bytes] = []

    # noinspection PyMethodMayBeStatic
    def _create_dock(self) -> Dock:
        """ Override for user-customized dock """
        return LockedDock()

    @property
    def delegate(self) -> Optional[DockerDelegate]:
        ref = self.__delegate_ref
        if ref is not None:
            return ref()

    @delegate.setter
    def delegate(self, keeper: DockerDelegate):
        self.__delegate_ref = None if keeper is None else weakref.ref(keeper)

    #
    #   Connection
    #

    @property
    def connection(self) -> Optional[Connection]:
        ref = self.__conn_ref
        if ref is not None:
            return ref()

    # Override
    def set_connection(self, connection: Optional[Connection]):
        # 1. replace with new connection
        old = self.connection
        if connection is None:
            self.__conn_ref = None
            self.__closed = True
        else:
            self.__conn_ref = weakref.ref(connection)
            self.__closed = False  # connection.closed
        # 2. close old connection
        if old is not None and old is not connection:
            old.close()

    #
    #   Flags
    #

    @property  # Override
    def closed(self) -> bool:
        if self.__closed is None:
            # initializing
            return False
        conn = self.connection
        return conn is None or conn.closed

    @property  # Override
    def alive(self) -> bool:
        conn = self.connection
        return conn is not None and conn.alive

    @property  # Override
    def status(self) -> DockerStatus:
        conn = self.connection
        state = None if conn is None else conn.state
        return status_from_state(state=state)

    def __str__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s>\n%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.connection, cname, mod)

    def __repr__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s>\n%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.connection, cname, mod)

    # Override
    def send_ship(self, ship: Departure) -> bool:
        return self.__dock.add_departure(ship=ship)

    # Override
    def process_received(self, data: bytes):
        # 1. get income ship from received data
        ships = self._get_arrivals(data=data)
        if ships is None or len(ships) == 0:
            # waiting for more data
            return None
        delegate = self.delegate
        for income in ships:
            # 2. check income ship for response
            income = self._check_arrival(ship=income)
            if income is None:
                # waiting for more fragment
                continue
            # 3. callback for processing income ship with completed data package
            if delegate is not None:
                delegate.docker_received(ship=income, docker=self)

    @abstractmethod
    def _get_arrivals(self, data: bytes) -> List[Arrival]:
        """
        Get income ships from received data

        :param data: received data
        :return income ships carrying data package/fragments
        """
        raise NotImplemented

    @abstractmethod
    def _check_arrival(self, ship: Arrival) -> Optional[Arrival]:
        """
        Check income ship for responding

        :param ship: income ship carrying data package/fragment/response
        :return income ship carrying completed data package
        """
        raise NotImplemented

    def _check_response(self, ship: Arrival) -> Optional[Departure]:
        """
        Check and remove linked departure ship with same SN (and page index for fragment)

        :param ship: income ship with SN
        """
        # check response for linked departure ship (same SN)
        linked = self.__dock.check_response(ship=ship)
        if linked is None:
            # linked departure task not found, or not finished yet
            return None
        # all fragments responded, task finished
        delegate = self.delegate
        if delegate is not None:
            delegate.docker_sent(ship=linked, docker=self)
        return linked

    def _assemble_arrival(self, ship: Arrival) -> Optional[Arrival]:
        """ Check received ship for completed package """
        return self.__dock.assemble_arrival(ship=ship)

    def _next_departure(self, now: float) -> Optional[Departure]:
        """ Get outgo ship from waiting queue """
        return self.__dock.next_departure(now=now)

    # Override
    def purge(self, now: float = 0) -> int:
        return self.__dock.purge(now=now)

    # Override
    def close(self):
        conn = self.connection
        if conn is not None:
            conn.close()

    #
    #  Processor
    #

    # Override
    def process(self) -> bool:
        # 1. get connection which is ready for sending data
        conn = self.connection
        if conn is None or not conn.alive:
            # waiting for connection
            return False
        # 2. get data waiting to be sent out
        outgo = self.__last_outgo
        fragments = self.__last_fragments
        if outgo is not None and len(fragments) > 0:
            # got remaining fragments from last outgo task
            self.__last_outgo = None
            self.__last_fragments = []
        else:
            # get next outgo task
            now = time.time()
            outgo = self._next_departure(now=now)
            if outgo is None:
                # nothing to do now, return false to let the thread have a rest
                return False
            elif outgo.get_status(now=now) == ShipStatus.FAILED:
                delegate = self.delegate
                if delegate is not None:
                    # callback for mission failed
                    error = TimeoutError('Request timeout')
                    delegate.docker_failed(error=error, ship=outgo, docker=self)
                # task timeout, return True to process next one
                return True
            else:
                # get fragments from outgo task
                fragments = outgo.fragments
                if len(fragments) == 0:
                    # all fragments of this task have been sent already
                    # return True to process next one
                    return True
        # 3. process fragments of outgo task
        index = 0
        sent = 0
        try:
            for fra in fragments:
                sent = conn.send(data=fra)
                if sent < len(fra):
                    # buffer overflow?
                    break
                else:
                    # assert sent == len(fra), 'length of fragment error: %d, %d' % (sent, len(fra))
                    index += 1
                    sent = 0  # clear counter
            if index < len(fragments):
                # task failed
                error = ConnectionError('only %d/%d fragments sent.' % (index, len(fragments)))
            else:
                # task done
                return True
        except Exception as e:
            # socket error, callback
            error = e
        # 4. remove sent fragments
        while index > 0:
            fragments.pop(0)
            index -= 1
        # remove partially sent data of next fragment
        if sent > 0:
            last = fragments.pop(0)
            fragments.insert(0, last[sent:])
        # 5. store remaining data
        self.__last_outgo = outgo
        self.__last_fragments = fragments
        # 6. callback for error
        delegate = self.delegate
        if delegate is not None:
            delegate.docker_error(error=error, ship=outgo, docker=self)
        # task error
        return False



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import time
import weakref
from typing import Optional

from ..types import SocketAddress, AddressPairObject
from ..fsm import Delegate as StateDelegate

from ..net import Hub
from ..net import Channel
from ..net import Connection, ConnectionState
from ..net import ConnectionDelegate
from ..net.state import StateMachine, StateOrder, TimedConnection


class BaseConnection(AddressPairObject, Connection, TimedConnection, StateDelegate):

    EXPIRES = 16  # seconds

    def __init__(self, remote: SocketAddress, local: Optional[SocketAddress]):
        super().__init__(remote=remote, local=local)
        self.__delegate_ref = None
        self.__channel_ref = None
        self.__closed = None
        # active times
        self.__last_sent_time = 0
        self.__last_received_time = 0
        # Finite State Machine
        self.__fsm: Optional[StateMachine] = None

    @property
    def delegate(self) -> ConnectionDelegate:
        """ Delegate for handling connection events """
        ref = self.__delegate_ref
        if ref is not None:
            return ref()

    @delegate.setter
    def delegate(self, gate: ConnectionDelegate):
        self.__delegate_ref = None if gate is None else weakref.ref(gate)

    #
    #   Channel
    #

    @property
    def channel(self) -> Optional[Channel]:
        ref = self.__channel_ref
        if ref is not None:
            return ref()

    def _set_channel(self, channel: Optional[Channel]):
        # 1. replace with new channel
        old = self.channel
        if channel is None:
            self.__channel_ref = None
            self.__closed = True
        else:
            self.__channel_ref = weakref.ref(channel)
            self.__closed = False  # channel.closed
        # 2. close old channel
        if old is not None and old is not channel:
            old.close()

    #
    #   State Machine
    #

    def _get_state_machine(self) -> Optional[StateMachine]:
        return self.__fsm

    # private
    def _set_state_machine(self, fsm: Optional[StateMachine]):
        # 1. replace with new machine
        old = self.__fsm
        self.__fsm = fsm
        # 2. stop old machine
        if old is not None and old is not fsm:
            old.stop()

    def _create_state_machine(self) -> StateMachine:
        fsm = StateMachine(connection=self)
        fsm.delegate = self
        return fsm

    #
    #   Flags
    #

    @property  # Override
    def closed(self) -> bool:
        if self.__closed is None:
            # initializing
            return False
        channel = self.channel
        return channel is None or channel.closed

    @property  # Override
    def bound(self) -> bool:
        channel = self.channel
        return channel is not None and channel.bound

    @property  # Override
    def connected(self) -> bool:
        channel = self.channel
        return channel is not None and channel.connected

    @property  # Override
    def alive(self) -> bool:
        # channel = self.channel
        # return channel is not None and channel.alive
        return (not self.closed) and (self.connected or self.bound)

    def __str__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s>\n%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.channel, cname, mod)

    def __repr__(self) -> str:
        mod = self.__module__
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s>\n%s\n</%s module="%s">'\
               % (cname, self._remote, self._local, self.channel, cname, mod)

    # Override
    def close(self):
        # stop state machine
        self._set_state_machine(fsm=None)
        # close channel
        self._set_channel(channel=None)

    def start(self, hub: Hub):
        # 1. get channel from hub
        self._open_channel(hub=hub)
        # 2. start state machine
        self._start_machine()

    def _start_machine(self):
        fsm = self._create_state_machine()
        self._set_state_machine(fsm=fsm)
        fsm.start()

    # protected
    def _open_channel(self, hub: Hub) -> Optional[Channel]:
        sock = hub.open(remote=self.remote_address, local=self.local_address)
        if sock is None:
            assert False, 'failed to open channel: remote=%s, local=%s' % (self.remote_address, self.local_address)
        else:
            self._set_channel(channel=sock)
        return sock

    #
    #   I/O
    #

    # Override
    def received(self, data: bytes):
        self.__last_received_time = time.time()  # update received time
        delegate = self.delegate
        if delegate is not None:
            delegate.connection_received(data=data, connection=self)

    def _send(self, data: bytes, target: SocketAddress) -> int:
        channel = self.channel
        if channel is None or not channel.alive:
            # raise socket.error('socket channel lost: %s' % channel)
            return -1
        sent = channel.send(data=data, target=target)
        if sent > 0:
            self.__last_sent_time = time.time()
        return sent

    # Override
    def send(self, data: bytes) -> int:
        # try to send data
        error = None
        sent = -1
        try:
            sent = self._send(data=data, target=self.remote_address)
            if sent < 0:  # == -1:
                raise socket.error('failed to send: %d byte(s) to %s' % (len(data), self.remote_address))
        except socket.error as e:
            error = e
            # socket error, close current channel
            self._set_channel(channel=None)
        # callback
        delegate = self.delegate
        if delegate is not None:
            if error is None:
                delegate.connection_sent(sent=sent, data=data, connection=self)
            else:
                delegate.connection_failed(error=error, data=data, connection=self)
        return sent

    #
    #   States
    #

    @property  # Override
    def state(self) -> Optional[ConnectionState]:
        fsm = self._get_state_machine()
        if fsm is not None:
            return fsm.current_state

    # Override
    def tick(self, now: float, elapsed: float):
        if self.__closed is None:
            # not initialized
            return
        fsm = self._get_state_machine()
        if fsm is not None:
            # drive state machine forward
            fsm.tick(now=now, elapsed=elapsed)

    #
    #   Timed Connection
    #

    @property  # Override
    def last_sent_time(self) -> float:
        return self.__last_sent_time

    @property  # Override
    def last_received_time(self) -> float:
        return self.__last_received_time

    # Override
    def is_sent_recently(self, now: float) -> bool:
        return now <= self.__last_sent_time + self.EXPIRES

    # Override
    def is_received_recently(self, now: float) -> bool:
        return now <= self.__last_received_time + self.EXPIRES

    # Override
    def is_not_received_long_time_ago(self, now: float) -> bool:
        return now > self.__last_received_time + (self.EXPIRES << 3)

    #
    #   StateDelegate
    #

    # Override
    def enter_state(self, state: Optional[ConnectionState], ctx: StateMachine, now: float):
        pass

    # Override
    def exit_state(self, state: Optional[ConnectionState], ctx: StateMachine, now: float):
        current = ctx.current_state
        if isinstance(current, ConnectionState):
            index = current.index
        else:
            assert current is None, 'unknown connection state: %s' % current
            index = -1
        # if current == 'ready'
        if index == StateOrder.READY:
            previous = -1 if state is None else state.index
            # if preparing == 'preparing'
            if previous == StateOrder.PREPARING:
                # connection state changed from 'preparing' to 'ready',
                # set times to expired soon.
                timestamp = time.time() - (self.EXPIRES >> 1)
                if self.__last_sent_time < timestamp:
                    self.__last_sent_time = timestamp
                if self.__last_received_time < timestamp:
                    self.__last_received_time = timestamp
        # callback
        delegate = self.delegate
        if delegate is not None:
            delegate.connection_state_changed(previous=state, current=current, connection=self)

    # Override
    def pause_state(self, state: Optional[ConnectionState], ctx: StateMachine, now: float):
        pass

    # Override
    def resume_state(self, state: Optional[ConnectionState], ctx: StateMachine, now: float):
        pass



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
from abc import ABC, abstractmethod
from typing import Optional, Union

from .connection import Connection
from .state import ConnectionState


class ConnectionDelegate(ABC):
    """ Connection Delegate """

    @abstractmethod
    def connection_state_changed(self, previous: Optional[ConnectionState], current: Optional[ConnectionState],
                                 connection: Connection):
        """
        Called when connection status is changed

        :param previous:   old state
        :param current:    new state
        :param connection: current connection
        """
        raise NotImplemented

    @abstractmethod
    def connection_received(self, data: bytes, connection: Connection):
        """
        Called when connection received data

        :param data:        received data package
        :param connection:  current connection
        """
        raise NotImplemented

    @abstractmethod
    def connection_sent(self, sent: int, data: bytes, connection: Connection):
        """
        Called after data sent via the connection

        :param sent:        length of sent bytes
        :param data:        sent data package
        :param connection:  current connection
        """
        raise NotImplemented

    @abstractmethod
    def connection_failed(self, error: Union[IOError, socket.error], data: bytes, connection: Connection):
        """
        Called when failed to send data via the connection

        :param error:       connection error
        :param data:        outgoing data package
        :param connection:  current connection
        """
        raise NotImplemented

    @abstractmethod
    def connection_error(self, error: Union[IOError, socket.error], connection: Connection):
        """
        Called when connection (receiving) error

        :param error:       connection error
        :param connection:  current connection
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC  # , abstractmethod

from .ticker import PrimeMetronome
from .machine import S, C, U, T
from .base import BaseMachine


# noinspection PyAbstractClass
class AutoMachine(BaseMachine[C, T, S], ABC):

    # @property  # Override
    # @abstractmethod
    # def context(self) -> C:
    #     """ machine itself """
    #     raise NotImplemented

    # Override
    def start(self):
        super().start()
        timer = PrimeMetronome()
        timer.add_ticker(ticker=self)

    # Override
    def stop(self):
        timer = PrimeMetronome()
        timer.remove_ticker(ticker=self)
        super().stop()

    # Override
    def pause(self):
        timer = PrimeMetronome()
        timer.remove_ticker(ticker=self)
        super().pause()

    # Override
    def resume(self):
        super().resume()
        timer = PrimeMetronome()
        timer.add_ticker(ticker=self)



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import threading
import time
import weakref
from abc import ABC, abstractmethod
from typing import Optional, List, Iterable, Union

from .types import SocketAddress, AddressPairMap
from .net import Connection, ConnectionDelegate, ConnectionState
from .net.state import StateOrder
from .port import Departure, Gate
from .port import Docker, DockerStatus, DockerDelegate
from .port.docker import status_from_state


class DockerPool(AddressPairMap[Docker]):

    # Override
    def set(self, item: Optional[Docker],
            remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Docker]:
        # 1. remove cached item
        cached = super().remove(item=item, remote=remote, local=local)
        if cached is not None and cached is not item:
            cached.close()
        # 2. set new item
        old = super().set(item=item, remote=remote, local=local)
        assert old is None, 'should not happen'
        return cached

    # Override
    def remove(self, item: Optional[Docker],
               remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Docker]:
        cached = super().remove(item=item, remote=remote, local=local)
        if cached is not None and cached is not item:
            cached.close()
        if item is not None:
            item.close()
        return cached


class StarGate(Gate, ConnectionDelegate, ABC):
    """
        Star Gate
        ~~~~~~~~~

        @abstract methods:
            - _create_docker(advance_party, remote_address, local_address)
            - _cache_advance_party(data, connection)
            - _clear_advance_party(connection)
    """

    def __init__(self, delegate: DockerDelegate):
        super().__init__()
        self.__delegate_ref = weakref.ref(delegate)
        self.__docker_pool = self._create_docker_pool()
        self.__lock = threading.Lock()

    # noinspection PyMethodMayBeStatic
    def _create_docker_pool(self):
        return DockerPool()

    @property
    def delegate(self) -> Optional[DockerDelegate]:
        return self.__delegate_ref()

    # Override
    def send_data(self, payload: Union[bytes, bytearray],
                  remote: SocketAddress, local: Optional[SocketAddress]) -> bool:
        worker = self._get_docker(remote=remote, local=local)
        if worker is None:
            # assert False, 'docker not found: %s -> %s' % (local, remote)
            return False
        elif not worker.alive:
            # assert False, 'docket not alive: %s -> %s' % (local, remote)
            return False
        return worker.send_data(payload=payload)

    # Override
    def send_ship(self, ship: Departure, remote: SocketAddress, local: Optional[SocketAddress]) -> bool:
        worker = self._get_docker(remote=remote, local=local)
        if worker is None:
            # assert False, 'docker not found: %s -> %s' % (local, remote)
            return False
        elif not worker.alive:
            # assert False, 'docket not alive: %s -> %s' % (local, remote)
            return False
        return worker.send_ship(ship=ship)

    #
    #   Docker
    #

    @abstractmethod
    def _create_docker(self, parties: List[bytes],
                       remote: SocketAddress, local: Optional[SocketAddress]) -> Docker:
        """
        create new docker for received data

        :param parties: advance party
        :param remote:  remote address
        :param local:   local address
        :return docker
        """
        raise NotImplemented

    def _all_dockers(self) -> Iterable[Docker]:
        """ get a copy of all dockers """
        return self.__docker_pool.items

    def _remove_docker(self, docker: Optional[Docker],
                       remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Docker]:
        """ remove cached docker """
        return self.__docker_pool.remove(item=docker, remote=remote, local=local)

    def _get_docker(self, remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Docker]:
        """ get cached docker """
        return self.__docker_pool.get(remote=remote, local=local)

    def _set_docker(self, docker: Docker,
                    remote: SocketAddress, local: Optional[SocketAddress]) -> Optional[Docker]:
        """ cache docker """
        return self.__docker_pool.set(item=docker, remote=remote, local=local)

    #
    #   Processor
    #

    # Override
    def process(self) -> bool:
        dockers = self._all_dockers()
        # 1. drive all dockers to process
        count = self._drive_dockers(dockers=dockers)
        # 2. cleanup for dockers
        self._cleanup_dockers(dockers=dockers)
        return count > 0

    # noinspection PyMethodMayBeStatic
    def _drive_dockers(self, dockers: Iterable[Docker]) -> int:
        count = 0
        for worker in dockers:
            if worker.process():
                count += 1  # it's busy
        return count

    def _cleanup_dockers(self, dockers: Iterable[Docker]):
        now = time.time()
        for worker in dockers:
            if worker.closed:
                # remove docker which connection lost
                self._remove_docker(docker=worker, remote=worker.remote_address, local=worker.local_address)
            else:
                # clear expired tasks
                worker.purge(now=now)

    def _heartbeat(self, connection: Connection):
        """ Send a heartbeat package('PING') to remote address """
        remote = connection.remote_address
        local = connection.local_address
        worker = self._get_docker(remote=remote, local=local)
        if worker is not None:
            worker.heartbeat()

    #
    #   Connection Delegate
    #

    # Override
    def connection_state_changed(self, previous: Optional[ConnectionState], current: Optional[ConnectionState],
                                 connection: Connection):
        # convert status
        s1 = status_from_state(state=previous)
        s2 = status_from_state(state=current)
        # 1. callback when status changed
        if s1 != s2:
            remote = connection.remote_address
            local = connection.local_address
            # try to get docker
            with self.__lock:
                old = self._get_docker(remote=remote, local=local)
                if old is None:
                    if s2 == DockerStatus.ERROR:
                        # connection closed and docker removed
                        return
                    # create & cache docker
                    worker = self._create_docker([], remote=remote, local=local)
                    self._set_docker(worker, remote=remote, local=local)
                else:
                    worker = old
            if old is None:
                # set connection for this docker
                worker.set_connection(connection)
            # NOTICE: if the previous state is null, the docker maybe not
            #         created yet, this situation means the docker status
            #         not changed too, so no need to callback here.
            delegate = self.delegate
            if delegate is not None:
                delegate.docker_status_changed(previous=s1, current=s2, docker=worker)
        # 2. heartbeat when connection expired
        index = -1 if current is None else current.index
        if index == StateOrder.EXPIRED:
            self._heartbeat(connection=connection)

    # Override
    def connection_received(self, data: bytes, connection: Connection):
        remote = connection.remote_address
        local = connection.local_address
        # try to get docker
        with self.__lock:
            old = self._get_docker(remote=remote, local=local)
            if old is None:
                # cache advance party for this connection
                party = self._cache_advance_party(data=data, connection=connection)
                assert party is not None and len(party) > 0, 'advance party error'
                # create & cache docker
                worker = self._create_docker(party, remote=remote, local=local)
                self._set_docker(worker, remote=remote, local=local)
            else:
                party = []
                worker = old
        if old is None:
            # set connection for this docker
            worker.set_connection(connection)
            # process advance parties one by one
            for item in party:
                worker.process_received(data=item)
            # remove advance party
            self._clear_advance_party(connection=connection)
        else:
            # docker exists, call docker.onReceived(data)
            worker.process_received(data=data)

    @abstractmethod
    def _cache_advance_party(self, data: bytes, connection: Connection) -> List[bytes]:
        """
        Cache the advance party before decide which docker to use

        :param data:        received data
        :param connection:  current connection
        :return all cached data
        """
        raise NotImplemented

    @abstractmethod
    def _clear_advance_party(self, connection: Connection):
        """
        Clear all advance parties after docker created

        :param connection:  current connection
        """
        raise NotImplemented

    # Override
    def connection_sent(self, sent: int, data: bytes, connection: Connection):
        # ignore event for sending success
        pass

    # Override
    def connection_failed(self, error: Union[IOError, socket.error], data: bytes, connection: Connection):
        # ignore event for sending failed
        pass

    # Override
    def connection_error(self, error: Union[IOError, socket.error], connection: Connection):
        # ignore event for receiving error
        pass



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Optional, Set, Iterable, MutableMapping

from .pair import SocketAddress


K = TypeVar('K')
V = TypeVar('V')


class PairMap(Generic[K, V], ABC):

    @property
    @abstractmethod
    def items(self) -> Iterable[V]:
        """ Get all mapped items """
        raise NotImplemented

    @abstractmethod
    def get(self, remote: Optional[K], local: Optional[K]) -> Optional[V]:
        """ Get item by key pair (remote, local) """
        raise NotImplemented

    @abstractmethod
    def set(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        """ Set item by key pair (remote, local) """
        raise NotImplemented

    @abstractmethod
    def remove(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        """ Remove mapping item by key pair (remote, local) """
        raise NotImplemented


# noinspection PyAbstractClass
class AbstractPairMap(PairMap[K, V], ABC):

    def __init__(self, default: K):
        super().__init__()
        self.__default = default  # default key
        self.__map: MutableMapping[K, MutableMapping[K, V]] = {}  # (K, K) => V

    # @property  # Override
    # def items(self) -> Iterable[V]:
    #     # Caveat: the iterator will keep a strong reference to
    #     # `item` in WeakSet until it is resumed or closed.
    #     return self.__map.items()

    # Override
    def get(self, remote: Optional[K], local: Optional[K]) -> Optional[V]:
        if remote is None:
            assert local is not None, 'local & remote addresses should not empty at the same time'
            key1 = local
            key2 = None
        else:
            key1 = remote
            key2 = local
        table = self.__map.get(key1)
        if table is None:
            return None
        if key2 is not None:
            # mapping: (remote, local) => Connection
            item = table.get(key2)
            if item is not None:
                return item
            # take any Connection connected to remote
            return table.get(self.__default)
        else:
            # mapping: (remote, None) => Connection
            # mapping: (local, None) => Connection
            item = table.get(self.__default)
            if item is not None:
                # take the item with empty key2
                return item
            # take any Connection connected to remote / bound to local
            for name in table:
                item = table.get(name)
                if item is not None:
                    return item

    # Override
    def set(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        # create indexes with key pair (remote, local)
        if remote is None:
            assert local is not None, 'local & remote addresses should not empty at the same time'
            key1 = local
            key2 = self.__default
        elif local is None:
            key1 = remote
            key2 = self.__default
        else:
            key1 = remote
            key2 = local
        table = self.__map.get(key1)
        if table is not None:
            if item is not None:
                old = table.get(key2)
                table[key2] = item
                return old
            else:
                return table.pop(key2, None)
        elif item is not None:
            table = weakref.WeakValueDictionary()
            table[key2] = item
            self.__map[key1] = table

    # Override
    def remove(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        # remove indexes with key pair (remote, local)
        if remote is None:
            assert local is not None, 'local & remote addresses should not empty at the same time'
            key1 = local
            key2 = self.__default
        elif local is None:
            key1 = remote
            key2 = self.__default
        else:
            key1 = remote
            key2 = local
        table = self.__map.get(key1)
        if table is None:
            return item
        old = table.pop(key2, None)
        if len(table) == 0:
            self.__map.pop(key1, None)
        return item if old is None else old


class HashPairMap(AbstractPairMap[K, V]):

    def __init__(self, default: K):
        super().__init__(default=default)
        self.__items: Set[V] = set()

    @property  # Override
    def items(self) -> Iterable[V]:
        return self.__items.copy()

    # Override
    def set(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        if item is not None:
            # the caller may create different values with same pair (remote, local)
            # so here we should try to remove it first to make sure it's clean
            self.__items.discard(item)
            # cache it
            self.__items.add(item)
        # create indexes
        old = super().set(item=item, remote=remote, local=local)
        # clear replaced value
        if old is not None and old != item:
            self.__items.discard(old)
        return old

    # Override
    def remove(self, item: Optional[V], remote: Optional[K], local: Optional[K]) -> Optional[V]:
        # remove indexes
        old = super().remove(item=item, remote=remote, local=local)
        if old is not None:
            self.__items.discard(old)
            # return old
        if item is not None and item != old:
            self.__items.discard(item)
            # return item
        return item if old is None else old


class AddressPairMap(HashPairMap[SocketAddress, V]):

    AnyAddress = ('0.0.0.0', 0)

    def __init__(self):
        super().__init__(default=self.AnyAddress)



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
from abc import ABC, abstractmethod
from typing import Optional, Iterable

from ..types import SocketAddress
from ..fsm import Processor

from .channel import disconnect_socket
from .channel import Channel
from .connection import Connection


class Hub(Processor, ABC):
    """ Connections & Channels Container """

    @abstractmethod
    def open(self, remote: Optional[SocketAddress], local: Optional[SocketAddress]) -> Optional[Channel]:
        """
        Open a channel with direction (remote, local)

        :param remote: remote address to connected
        :param local:  local address to bound
        :return: None on socket error
        """
        raise NotImplemented

    @abstractmethod
    def connect(self, remote: SocketAddress, local: Optional[SocketAddress] = None) -> Optional[Connection]:
        """
        Get connection with direction (remote, local)

        :param remote: remote address
        :param local:  local address
        :return: None on channel closed
        """
        raise NotImplemented

    #
    #   Local SocketAddress
    #

    @classmethod
    def host_name(cls) -> str:
        return socket.gethostname()

    @classmethod
    def addr_info(cls):  # -> List[Tuple[Union[AddressFamily, int], Union[SocketKind, int], int, str, Tuple[Any, ...]]]
        host = socket.gethostname()
        if host is not None:
            try:
                return socket.getaddrinfo(host, None)
            except socket.error as error:
                print('[NET] failed to get address info: %s' % error)
                # traceback.print_exc()
                return []

    @classmethod
    def inet_addresses(cls) -> Iterable[str]:
        addresses = set()
        info = cls.addr_info()
        for item in info:
            addresses.add(item[4][0])
        return addresses

    @classmethod
    def inet_address(cls) -> Optional[str]:
        # get from addr info
        info = cls.addr_info()
        for item in info:
            ip = item[4][0]
            if ':' not in ip and '127.0.0.1' != ip:
                return ip
        # get from UDP socket
        sock = None
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            remote = ('8.8.8.8', 8888)
            sock.connect(remote)
            ip = sock.getsockname()[0]
        finally:
            if sock is not None:
                disconnect_socket(sock=sock)
        return ip



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union


# Addresses can be either tuples of varying lengths (AF_INET, AF_INET6,
# AF_NETLINK, AF_TIPC) or strings (AF_UNIX).
SocketAddress = Union[tuple, str]
# TODO: Most methods allow bytes as address objects


class AddressPairObject:

    def __init__(self, remote: Optional[SocketAddress], local: Optional[SocketAddress]):
        super().__init__()
        self._remote = remote
        self._local = local

    @property
    def remote_address(self) -> Optional[SocketAddress]:
        return self._remote

    @property
    def local_address(self) -> Optional[SocketAddress]:
        return self._local

    def __str__(self) -> str:
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s />' % (cname, self.remote_address, self.local_address)

    def __repr__(self) -> str:
        cname = self.__class__.__name__
        return '<%s: remote=%s, local=%s />' % (cname, self.remote_address, self.local_address)

    def __hash__(self) -> int:
        # same algorithm as Pair::hashCode()
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # remote's hashCode is multiplied by an arbitrary prime number (13)
        # in order to make sure there is a difference in the hashCode between
        # these two parameters:
        #     remote: a  local: aa
        #     local: aa  remote: a
        remote = self._remote
        local = self._local
        if remote is None:
            return hash(local)  # if local is not None else 0
        elif local is None:
            return hash(remote) * 13
        else:
            return hash(remote) * 13 + hash(local)

    def __eq__(self, other) -> bool:
        if other is None:
            return self._remote is None and self._local is None
        elif other is self:
            return True
        elif isinstance(other, AddressPairObject):
            return self._remote == other._remote and self._local == other._local

    def __ne__(self, other) -> bool:
        if other is None:
            return self._remote is not None or self._local is not None
        elif other is self:
            return False
        elif isinstance(other, AddressPairObject):
            return self._remote != other._remote or self._local != other._local
        else:
            return True



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
import weakref
from typing import Optional

from ..types import SocketAddress
from ..fsm import Daemon
from ..net import Hub

from .base_conn import BaseConnection


class ActiveConnection(BaseConnection):
    """ Active connection for client """

    def __init__(self, remote: SocketAddress, local: Optional[SocketAddress]):
        super().__init__(remote=remote, local=local)
        self.__hub_ref = None
        self.__daemon = Daemon(target=self.run)

    @property
    def hub(self) -> Hub:
        return self.__hub_ref()

    @property  # Override
    def closed(self) -> bool:
        return self._get_state_machine() is None

    # Override
    def start(self, hub: Hub):
        self.__hub_ref = weakref.ref(hub)
        # 1. start state machine
        self._start_machine()
        # 2. start a background thread to check channel
        self.__daemon.start()

    # protected
    def run(self):
        expired = 0
        last_time = 0
        interval = 16
        while not self.closed:
            time.sleep(1.0)
            #
            #  1. check time interval
            #
            now = time.time()
            if now < (last_time + interval):
                continue
            last_time = now
            if interval < 256:
                interval *= 2
            #
            #  2. check socket channel
            #
            sock = self.channel
            if sock is None or sock.closed:
                # get new socket channel via hub
                hub = self.hub
                if hub is None:
                    # assert False, 'hub lost'
                    break
                sock = self._open_channel(hub=hub)
                if sock is not None:
                    # connect timeout after 2 minutes
                    expired = now + 128
            elif sock.alive:
                # socket channel is normal
                interval = 16
            elif 0 < expired < now:
                # connect timeout
                sock.close()



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional

from ..types import SocketAddress
from ..fsm import Ticker


class Connection(Ticker, ABC):

    #
    #   Flags
    #

    @property
    @abstractmethod
    def closed(self) -> bool:
        """ not is_open() """
        raise NotImplemented

    @property
    @abstractmethod
    def bound(self) -> bool:
        """ is_bound() """
        raise NotImplemented

    @property
    @abstractmethod
    def connected(self) -> bool:
        """ is_connected() """
        raise NotImplemented

    @property
    @abstractmethod
    def alive(self) -> bool:
        """ is_opened() and (is_connected() or is_bound()) """
        raise NotImplemented

    @property
    @abstractmethod
    def local_address(self) -> Optional[SocketAddress]:  # (str, int)
        raise NotImplemented

    @property
    @abstractmethod
    def remote_address(self) -> Optional[SocketAddress]:  # (str, int)
        raise NotImplemented

    @property
    @abstractmethod
    def state(self):  # -> Optional[ConnectionState]:
        """ Get connection state """
        raise NotImplemented

    @abstractmethod
    def start(self, hub):
        """
        Get channel from hub

        :param hub: connection hub
        """
        raise NotImplemented

    @abstractmethod
    def send(self, data: bytes) -> int:
        """
        Send data

        :param data: outgo data package
        :return: count of bytes sent, probably zero when it's non-blocking mode
        """
        raise NotImplemented

    @abstractmethod
    def received(self, data: bytes):
        """
        Call on received data for processing

        :param data: received data
        """
        raise NotImplemented

    @abstractmethod
    def close(self):
        """ Close the connection """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
import weakref
from abc import ABC
from typing import Optional, Any, Set, Dict, MutableMapping

from .port import ShipStatus
from .port import Arrival


# noinspection PyAbstractClass
class ArrivalShip(Arrival, ABC):

    # Arrival task will be expired after 5 minutes
    # if still not completed
    EXPIRES = 300  # seconds

    def __init__(self, now: float = 0):
        super().__init__()
        # last received time (timestamp in seconds)
        if now <= 0:
            now = time.time()
        self.__expired = now + self.EXPIRES

    # Override
    def touch(self, now: float):
        # update expired time
        self.__expired = now + self.EXPIRES

    # Override
    def get_status(self, now: float) -> ShipStatus:
        if now > self.__expired:
            return ShipStatus.EXPIRED
        else:
            return ShipStatus.ASSEMBLING


class ArrivalHall:
    """ Memory cache for Arrivals """

    def __init__(self):
        super().__init__()
        self.__arrivals: Set[Arrival] = set()
        self.__map: MutableMapping[Any, Arrival] = weakref.WeakValueDictionary()  # sn => Arrival
        self.__finished_times: Dict[Any, float] = {}                              # sn => timestamp

    def assemble_arrival(self, ship: Arrival) -> Optional[Arrival]:
        """
        Check received ship for completed package

        :param ship: income ship carrying data package/fragment
        :return a ship carrying completed data package
        """
        # 1. check ship ID (SN)
        sn = ship.sn
        if sn is None:
            # separated package ship must have SN for assembling
            # we consider it to be a ship carrying a whole package here
            return ship
        # 2. check cached ship
        cached = self.__map.get(sn)
        if cached is None:
            # check whether the task as already finished
            timestamp = self.__finished_times.get(sn, 0)
            if timestamp > 0:
                # task already finished
                return None
            # 3. new arrival, try assembling to check whether a fragment
            completed = ship.assemble(ship=ship)
            if completed is None:
                # it's a fragment, waiting for more fragments
                self.__arrivals.add(ship)
                self.__map[sn] = ship
                # ship.touch(now=time.time())
            # else, it's a completed package
        else:
            # 3. cached ship found, try assembling (insert as fragment)
            #    to check whether all fragments received
            completed = cached.assemble(ship=ship)
            if completed is None:
                # it's not completed yet, update expired time
                # and wait for more fragments.
                cached.touch(now=time.time())
            else:
                # all fragments received, remove cached ship
                self.__arrivals.discard(cached)
                self.__map.pop(sn, None)
                # mark finished time
                self.__finished_times[sn] = time.time()
        return completed

    def purge(self, now: float = 0) -> int:
        """ Clear all expired tasks """
        if now <= 0:
            now = time.time()
        count = 0
        # 1. seeking expired tasks
        arrivals = set(self.__arrivals)
        for ship in arrivals:
            if ship.get_status(now=now) == ShipStatus.EXPIRED:
                # task expired
                self.__arrivals.discard(ship)
                # remove mapping with SN
                sn = ship.sn
                if sn is not None:
                    self.__map.pop(sn, None)
                    # TODO: callback?
                count += 1
        # 2. seeking neglected finished times
        ago = now - 3600
        keys = set(self.__finished_times.keys())
        for sn in keys:
            when = self.__finished_times.get(sn)
            if when is None or when < ago:
                # long time ago
                self.__finished_times.pop(sn, None)
        return count



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2022 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2022 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod

from .ship import Arrival, Departure
from .docker import Docker, DockerStatus


class DockerDelegate(ABC):

    @abstractmethod
    def docker_received(self, ship: Arrival, docker: Docker):
        """
        Callback when new package received

        :param ship:    income data package container
        :param docker:  connection docker
        """
        raise NotImplemented

    @abstractmethod
    def docker_sent(self, ship: Departure, docker: Docker):
        """
        Callback when package sent

        :param ship:    outgo data package container
        :param docker:  connection connection
        """
        raise NotImplemented

    @abstractmethod
    def docker_failed(self, error: OSError, ship: Departure, docker: Docker):
        """
        Callback when failed to send package

        :param error:   error message
        :param ship:    outgo data package container
        :param docker:  connection docker
        """
        raise NotImplemented

    @abstractmethod
    def docker_error(self, error: OSError, ship: Departure, docker: Docker):
        """
        Callback when connection error

        :param error:   error message
        :param ship:    outgo data package container
        :param docker:  connection docker
        """
        raise NotImplemented

    @abstractmethod
    def docker_status_changed(self, previous: DockerStatus, current: DockerStatus, docker: Docker):
        """
        Callback when connection status changed

        :param previous: old status
        :param current:  new status
        :param docker:   connection docker
        """
        raise NotImplemented



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Star Trek
    ~~~~~~~~~

    Interstellar Transport
"""

from setuptools import setup, find_packages

__version__ = '1.1.0'
__author__ = 'Albert Moky'
__contact__ = 'albert.moky@gmail.com'

with open('README.md', 'r') as fh:
    readme = fh.read()

setup(
    name='startrek',
    version=__version__,
    url='https://github.com/moky/wormhole/',
    license='MIT',
    author=__author__,
    author_email=__contact__,
    description='Interstellar Transport',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
    ]
)
